"""Files in this directory are generated by running:

    python -m dagster_cloud.pex.grpc.compile
"""

from . import multi_pex_api_pb2 as multi_pex_api_pb2
from .multi_pex_api_pb2_grpc import (
    MultiPexApiServicer as MultiPexApiServicer,
    MultiPexApiStub as MultiPexApiStub,
)
